import re
from functools import lru_cache
from io import StringIO
from textwrap import TextWrapper
from typing import Set, Optional, Dict

import requests


@lru_cache
def all_available_apis() -> Dict[str, dict]:
    """
    returns directory descriptions of all available APIs. The key is
    the api id, which consists of the name and version concatenate with a colon.
    """
    response = requests.get("https://www.googleapis.com/discovery/v1/apis")
    if response.status_code != 200:
        raise ValueError(
            f"failed to retrieve API documents, {response.status_code}, {response.text}"
        )

    return {a["id"]: a for a in response.json()["items"]}


def all_available_versions_of(api_id: str) -> [str]:
    """
    returns all versions of associated with the api `api_id`.
    """
    name = api_id.split(":")[0]
    return list(
        map(
            lambda i: i.split(":")[-1],
            filter(lambda i: i.startswith(name + ":"), all_available_apis().keys()),
        )
    )


def all_available_api_names() -> set[str]:
    """
    returns all available api names.
    """
    return set(map(lambda i: i.split(":")[0], all_available_apis().keys()))


def all_available_api_ids() -> Set[str]:
    """
    returns all available api ids.
    """
    return set(list(all_available_apis().keys()))


class SchemaTypeDefinition(dict):
    """
    represents a JSON schema type definition.
    """

    def __init__(self, other: dict):
        super().__init__()
        self.update(other)

    @property
    def is_resource_ref(self) -> re.Match:
        """
        Returns the Match object if this type represents a google cloud platform resource reference.
        """
        return (
            re.search(
                r"URL\s+referring\s+to\s+an?\s+([^\s]+)\s+",
                self.get("description", ""),
                re.IGNORECASE,
            )
            if self.get("type", "") == "string"
            else None
        )

    @property
    def is_output_only(self) -> bool:
        return "[output only]" in self.get("description", "").lower()

    @property
    def is_input_only(self) -> bool:
        return "@inputonly" in self.get("description", "").lower()

    @property
    def is_deprecated(self) -> bool:
        return "deprecated" in self.get("description", "").lower()

    @property
    def is_required(self) -> bool:
        return self.get("description", "").lower().startswith("required.")

    @property
    def value_type(self) -> str:
        return self.get("type")

    @property
    def value_format(self) -> str:
        return self.get("format", "")

    @property
    def kind(self) -> Optional[str]:
        return self.get("properties", {}).get("kind", {}).get("default")

    @property
    def properties(self) -> Dict[str, "SchemaTypeDefinition"]:
        return {
            name: SchemaTypeDefinition(definition)
            for name, definition in self.get("properties", {}).items()
        }

    @property
    def description(self) -> str:
        writer = StringIO()
        wrapper = TextWrapper(width=72, break_long_words=False, break_on_hyphens=False)
        description = self.get("description", "")
        if description:
            writer.write("\n".join(wrapper.wrap(description)))

        descriptions = self.get("enumDescriptions", [])
        if descriptions:
            if writer.tell() != 0:
                writer.write("\n")
            writer.write("The possible values are:\n")
            for i, enum_value in enumerate(self.get("enum", [])):
                enum_description = descriptions[i] if i < len(descriptions) else ""
                if not enum_description:
                    continue
                wrapper.subsequent_indent = " " * (len(enum_value) + 5)
                writer.write("\n")
                writer.write(
                    "\n".join(wrapper.wrap(f"* `{enum_value}`: {enum_description}"))
                )
                writer.write("\n")
        return writer.getvalue()


class SchemaMethodDefinition(dict):
    def __init__(self, other: dict):
        super().__init__()
        self.update(other)

    @property
    def flat_path(self) -> str:
        """
        the path for the operation
        """
        return self.get("flatPath")

    @property
    def request(self) -> SchemaTypeDefinition:
        """
        the schema definition of the request
        """
        return SchemaTypeDefinition(self.get("request"))

    @property
    def parameters(self) -> Dict[str, "SchemaTypeDefinition"]:
        return {
            name: SchemaTypeDefinition(definition)
            for name, definition in self.get("parameters", {}).items()
        }


class SchemaResourceDefinition(dict):
    def __init__(self, resource_name: str, other: dict):
        super().__init__()
        self.resource_name = resource_name
        self.update(other)

    @property
    def methods(self) -> dict[SchemaMethodDefinition]:
        return {
            name: SchemaMethodDefinition(value)
            for name, value in self.get("methods", {}).items()
        }

    def get_insert_or_create_method(self) -> SchemaMethodDefinition:
        insert_definition = self.methods.get("insert", self.methods.get("create"))
        if not insert_definition:
            raise ValueError(
                f"No insert or create method found on resource {self.resource_name}"
            )
        return SchemaMethodDefinition(insert_definition)


class APIMetaData:
    @lru_cache
    def __init__(self, api_id: str):
        self.api_id = api_id
        self.api_name, self.api_version = api_id.split(":")
        self.document = {}

    @staticmethod
    @lru_cache
    def load(api_id: str) -> "APIMetaData":
        if api_id not in all_available_apis():
            raise ValueError(
                f"invalid api id: {api_id}, available versions are: {all_available_versions_of(api_id)}"
            )

        api = all_available_apis()[api_id]
        response = requests.get(api["discoveryRestUrl"])
        if response.status_code != 200:
            raise ValueError(
                f'failed to retrieve metadata from {api["discoveryRestUrl"]}, {response.text}'
            )
        result = APIMetaData(api_id)
        result.document = response.json()
        return result

    def get_schema_type_definition(self, type_name: str) -> SchemaTypeDefinition:
        result = self.document.get("schemas", {}).get(type_name)
        if not result:
            raise ValueError(
                f"no type {type_name} defined in schema for api {self.api_id}"
            )
        return SchemaTypeDefinition(result)

    def get_resource_definition(self, resource_name: str) -> SchemaResourceDefinition:
        legacy = self.document.get("resources", {})
        project_level = (
            legacy.get("projects", {})
            .get("resources", {})
            .get("locations", {})
            .get("resources", {})
        )
        org_level = legacy.get("organizations", {}).get("resources", {})

        if resource_name in org_level:
            return SchemaResourceDefinition(resource_name, org_level[resource_name])
        elif resource_name in project_level:
            return SchemaResourceDefinition(resource_name, project_level[resource_name])
        elif resource_name in legacy:
            return SchemaResourceDefinition(resource_name, legacy[resource_name])
        else:
            names = (
                set(legacy.keys()).union(org_level.keys()).union(project_level.keys())
            )
            raise ValueError(
                f"resource {resource_name} not found, available resources {', '.join(names)}"
            )



import os
import logging
from magic_module_skaffolder.skaffolder import main

if __name__ == "__main__":
    logging.basicConfig(
        level=os.getenv("LOG_LEVEL", "INFO"), format="%(levelname)s: %(message)s"
    )
    main()



import os
import logging
import re
from pathlib import Path
from typing import Optional, Dict, Any

import click
from pattern.text.en import singularize
from ruamel.yaml import YAML
from ruamel.yaml.scalarstring import LiteralScalarString

from magic_module_skaffolder.api_descriptions import (
    APIMetaData,
    SchemaTypeDefinition,
)
from magic_module_skaffolder.magic_module import Product, Field, Resource

yaml = YAML(typ="rt")
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.preserve_quotes = True
yaml.width = 255


class Skaffolder:
    """
    generates or updates a Magic Module Resource definition with the metadata of the
    Google Cloud APIs.
    """

    def __init__(self):
        pass

    def create_magic_module_field(
        self,
        api: APIMetaData,
        name: Optional[str],
        type_definition: SchemaTypeDefinition,
    ) -> Field:
        """
        creates a Magic Module field definition from a JSON schema type definition,
        :param api: the metadata of the api
        :param name:  the name of the field if any
        :param type_definition: the json schema definition of the field
        :return: a Magic Module field definition
        """
        result: Dict[str, Any] = {"name": name} if name else {}
        description = type_definition.description
        if description:
            result["description"] = (
                LiteralScalarString(description)
                if description.count("\n") >= 1
                else description
            )

        if type_definition.is_output_only:
            result["output"] = True

        if type_definition.is_input_only:
            result["ignore_read"] = True

        if type_definition.is_required:
            result["required"] = True

        value_type = type_definition.value_type
        value_format = type_definition.value_format
        target_type = (
            {
                "double": "Api::Type::Double",
                "float": "Api::Type::Double",
                "int32": "Api::Type::Integer",
                "int64": "Api::Type::Integer",
                "uint32": "Api::Type::Integer",
                "uint64": "Api::Type::Integer",
            }.get(value_format)
            if value_type == "string"
            else None
        )

        if not value_type:
            ref = type_definition.get("$ref")
            if not ref:
                raise ValueError(f"no type nor $ref found in {name}")

            referenced_type = api.get_schema_type_definition(ref)
            result = self.create_magic_module_field(api, name, referenced_type)

        elif value_type == "integer":
            result["type"] = "Api::Type::Integer"
        elif value_type == "number":
            result["type"] = "Api::Type::Double"
        elif value_type == "string":
            if "enum" in type_definition:
                result["type"] = "Api::Type::Enum"
                result["values"] = [f":{value}" for value in type_definition["enum"]]
            elif target_type:
                result["type"] = target_type
            elif value_format == "byte" and name == "fingerprint":
                result["type"] = "Api::Type::Fingerprint"
            elif "in RFC3339 text format" in description or (
                name and "timestamp" in name.lower()
            ):
                result["type"] = "Api::Type::Time"
            elif type_definition.is_resource_ref:
                # see https://googlecloudplatform.github.io/magic-modules/docs/how-to/add-mmv1-resource/#resourceref
                resource_reference = type_definition.is_resource_ref.group(1)
                (
                    resource_reference_name,
                    resource_reference_type,
                ) = resource_reference.split(".", 1)
                result["resource"] = resource_reference_type
                result["imports"] = (
                    "selfLink" if resource_reference_name == "compute" else "name"
                )
                result["type"] = "Api::Type::ResourceRef"
            else:
                result["type"] = "Api::Type::String"
        elif value_type == "boolean":
            result["type"] = "Api::Type::Boolean"
        elif value_type == "array":
            result["type"] = "Api::Type::Array"
            result["item_type"] = self.create_magic_module_field(
                api, None, SchemaTypeDefinition(type_definition["items"])
            )
        elif value_type == "object":
            if "properties" not in type_definition:
                if (
                    type_definition.get("additionalProperties", {}).get("type")
                    == "string"
                ):
                    result["type"] = "Api::Type::KeyValuePairs"
                else:
                    raise ValueError("object type is missing properties")
            else:
                result["type"] = "Api::Type::NestedObject"
                properties = []
                for (
                    property_name,
                    property_definition,
                ) in type_definition.properties.items():
                    properties.append(
                        self.create_magic_module_field(
                            api, property_name, property_definition
                        )
                    )
                if len(properties) == 1:
                    properties[0]["required"] = True
                result["properties"] = properties
        else:
            raise ValueError(f"unexpected type {value_type}")

        return Field.create(result)

    def generate_magic_module_resource_properties(
        self, api, resource_name, type_name
    ) -> (dict, bool):
        """
        generates the properties for the magic module resource definition. Unfamiliarity with
        the magic modules and their mapping from api definition, makes this a best effort
        approach.
        """
        result = {"name": type_name}
        type_definition = api.get_schema_type_definition(type_name)
        if type_definition.kind:
            result["kind"] = type_definition.kind

        resource_definition = api.get_resource_definition(resource_name)
        method = resource_definition.get_insert_or_create_method()

        def parameter_camel_case_to_snake_case(name: str):
            name = name.removesuffix("Id")
            name = singularize(name)
            name = re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()
            return name

        def add_base_url():
            """
            the base url is without the version prefix and appears
            to have all variable references in the definition replaced
            with a snake case, singular variable. eg. {projectsId} -> {{project}}.
            :return:
            """
            base_url = (
                method.flat_path[len(api.api_version) + 1 :]
                if method.flat_path.startswith(api.api_version + "/")
                else method.flat_path
            )

            def _singular_snake_case(match: re.Match) -> str:
                return "{{" + parameter_camel_case_to_snake_case(match.group(1)) + "}}"

            base_url = re.sub(r"{([^}]*)}", _singular_snake_case, base_url)
            result["base_url"] = base_url

        def add_self_link():
            if "name" in type_definition.properties.keys():
                result["self_link"] = result["base_url"] + "/{{name}}"

        def add_update_verb():
            patch_method = resource_definition.methods.get("patch")
            if patch_method:
                result["update_verb"] = "PATCH"
                if "updateMask" in patch_method.parameters:
                    result["update_mask"] = True

        def add_create_link():
            id_name = type_name[0].lower() + type_name[1:] + "Id"
            if (
                id_name not in method.parameters
                and "name" not in type_definition.properties
            ):
                return

            result["import_format"] = [result["base_url"] + "/{{name}}"]

            query_parameters = {}
            parameters = []
            readable_name = parameter_camel_case_to_snake_case(type_name).replace(
                "_", " "
            )
            for name, value in method.parameters.items():
                if name == id_name:
                    parameter = self.create_magic_module_field(
                        api, "name", type_definition.properties["name"]
                    )
                    parameter[
                        "description"
                    ] = f"A user-defined name which uniquely identifies a {readable_name}."
                    parameter["url_param_only"] = True
                    parameters.append(parameter)
                    if value.get("location") == "query":
                        query_parameters[name] = "{{name}}"
                elif name == "parent":
                    parameter = self.create_magic_module_field(api, "location", value)
                    parameter["default"] = "global"
                    parameter["url_param_only"] = True
                    parameter["description"] = f"the location of the {readable_name}."
                    if value.get("location") == "query":
                        query_parameters[name] = "{{location}}"

                    parameters.append(parameter)
                elif name == "updateMask":
                    continue
                else:
                    parameter = self.create_magic_module_field(
                        api, parameter_camel_case_to_snake_case(name), value
                    )
                    if value.get("location") == "query":
                        query_parameters[name] = (
                            "{{" + parameter_camel_case_to_snake_case(name) + "}}"
                        )
                    parameters.append(parameter)

            if query_parameters:
                result["create_link"] = (
                    result["base_url"]
                    + "/?"
                    + "&".join(
                        map(lambda q: f"{q[0]}={q[1]}", query_parameters.items())
                    )
                )
            if parameters:
                result["parameters"] = parameters

        add_base_url()
        add_self_link()
        add_update_verb()
        add_create_link()
        return result

    def create_magic_module_resource(
        self, api: APIMetaData, resource_name: str, type_name: str
    ) -> Resource:
        """
        Creates a magic module resource definition for the type `type_name` of the `resource_name`
        in the `api`.
        """

        type_definition = api.get_schema_type_definition(type_name)
        properties = self.generate_magic_module_resource_properties(
            api, resource_name, type_name
        )
        result = Resource(properties)

        result.update(self.create_magic_module_field(api, None, type_definition))
        if any(filter(lambda p: p.get("name") == "name", result.get("parameters", []))):
            result["properties"] = list(
                filter(lambda p: p.get("name") != "name", result["properties"])
            )
        return result


@click.group()
def main():
    """
    generates or updates a Magic Module Resource definitions with the metadata of the
    Google Cloud APIs.

    When generating the definitions, it will first use the `ga` interface, and supplement the
    definition with the `beta` interface, so that it can determine which fields are only
    available on the `beta` interface.

    The resulting definitions are not perfect and may need some polishing. Most properties required
    by the magic module are derived from the free text description field. So, please do check
    the result.

    Existing field definitions are not overwritten, so once inspected and correct you can
    rerun the merge operation as often as you want. Note that fields of a resource which do not
    exist in the API are removed.
    """
    pass


@main.command()
@click.option(
    "--product-directory",
    required=True,
    type=click.Path(file_okay=False, exists=True),
    help="the Magic Module product directory the resource definitions belong to",
)
@click.argument("resource", nargs=-1, required=True)
def generate(
    product_directory: str,
    resource: tuple[str],
):
    """
    generates a resource definition for the specified resources. It will derive the type from
    the resource name. So the resource name `backendServices` will have type `BackendService`
    as its resource type definition.
    """
    product_definition_file = Path(product_directory).joinpath("product.yaml")
    if not product_definition_file.exists():
        click.echo(
            f"'{product_directory} is not a magic module product directory",
            err=True,
        )

    product_definition = Product.load(product_definition_file)
    updater = Skaffolder()
    for resource_name in resource:
        type_name = (
            singularize(resource_name)[0].upper() + singularize(resource_name)[1:]
        )
        existing = None
        for provider_version, api_id in product_definition.get_api_ids():
            api = APIMetaData.load(api_id)
            defined = updater.create_magic_module_resource(
                api, resource_name, type_name
            )
            if existing:
                Resource.merge_resources(existing, defined, provider_version)
            else:
                existing = defined

        output_file = Path(product_directory).joinpath(f"{type_name}.yaml")
        logging.info("Writing to definition of %s to %s", type_name, output_file)
        with open(output_file, "w") as file:
            yaml.dump(existing, file)


@main.command()
@click.option(
    "--resource-file",
    type=click.Path(dir_okay=False, exists=True),
    required=True,
    help="containing the magic module resource definition",
)
@click.option(
    "--inplace",
    is_flag=True,
    default=False,
    required=False,
    help="update the resource definition file",
)
def update(resource_file: str, inplace: bool):
    """
    updates the resource definition in the resource-file to match the latest API specification.
    """
    product_definition_file = Path(resource_file).parent.joinpath("product.yaml")
    if not product_definition_file.exists():
        click.echo(
            f"magic module product definition file '{product_definition_file} not found",
            err=True,
        )

    product_definition = Product.load(product_definition_file)

    existing: Optional[Resource] = None
    try:
        existing = Resource.load(resource_file)
    except ValueError as error:
        click.echo(str(error), err=True)

    type_name = existing["name"]
    api_name, _ = existing["kind"].split("#")
    resource_name = existing["base_url"].split("/")[-1]

    updater = Skaffolder()

    for provider_version, api_id in product_definition.get_api_ids():
        api = APIMetaData.load(api_id)
        defined = updater.create_magic_module_resource(api, resource_name, type_name)
        Resource.merge_resources(existing, defined, provider_version)

    if inplace:
        with open(resource_file, "w") as file:
            yaml.dump(existing, file)
    else:
        yaml.dump(existing, os.stdout)



import logging
import re
from pathlib import Path
from typing import Optional, Generator, Union

from ruamel.yaml.comments import CommentedMap
from ruamel.yaml.representer import RoundTripRepresenter

from magic_module_skaffolder.yaml import yaml

_api_url_pattern = re.compile(r"https:\/\/(?P<name>[^\.]*)\.googleapis\.com\/.*")


class Product(CommentedMap):
    def __init__(self, value: CommentedMap):
        super().__init__()
        if value.tag.value != "!ruby/object:Api::Product":
            raise ValueError("value is not a magic module product definition")
        self.value = value

    @staticmethod
    def load(path: Path):
        with open(path, "r") as file:
            return Product(yaml.load(file))

    def get_api_id(self, version_name: str) -> Optional[str]:
        assert version_name in ["beta", "ga"]

        version = next(
            filter(lambda v: v["name"] == version_name, self.value.get("versions", [])),
            None,
        )
        if not version:
            return None

        base_url = version["base_url"]
        match = _api_url_pattern.fullmatch(base_url)
        if not match:
            raise ValueError(
                f"{base_url} does not match the expected google cloud platform API url"
            )

        name = match.group("name")
        version = base_url.strip("/").split("/")[-1]
        return f"{name}:{version}"

    def get_api_ids(self) -> Generator:
        """
        returns the available provider version name and associated api id.
        """
        for name in ["ga", "beta"]:
            api_id = self.get_api_id(name)
            if api_id:
                yield name, api_id


@yaml.register_class
class Field(CommentedMap):
    """
    represents a Magic Module Field definition.
    """

    def __init__(self, tag):
        super().__init__(self)
        self.yaml_set_tag(tag)

    @staticmethod
    def create(item: Union[dict, str]) -> "Field":
        if isinstance(item, str):
            if item.startswith("Api::Type::"):
                result = Field(f"!ruby/object:{item}")
                return result
            else:
                raise ValueError(
                    "cannot create a field from a string which is not an Api::Type name"
                )

        tag = item.pop("type", None)
        if tag:
            result = Field(f"!ruby/object:{tag}" if tag else "")
            result.update(item)
        else:
            result = item
        return result

    @staticmethod
    def merge_fields(
        existing: Union["Resource", "Field", dict, str],
        defined: Union["Resource", "Field", dict, str],
        path: [str],
        provider_version: str,
    ):
        """
        merges the fields from the `defined` into `existing` field definition.
        """
        assert existing is not None
        assert defined is not None
        assert provider_version in ["ga", "beta"]

        if isinstance(existing, str):
            if existing.startswith("Api::Type::"):
                result = Field(f"!ruby/object:{existing}")
                return result
            else:
                raise ValueError(
                    "cannot create a field from a string which is not an Api::Type name"
                )

        existing_name = existing.get("api_name", existing.get("name", ""))
        defined_name = defined.get("api_name", defined.get("name", ""))

        if defined_name != existing_name:
            logging.warning(
                'mismatch in field name %s: expected "%s" defined "%s"',
                ".".join(path),
                existing_name,
                defined_name,
            )
        path = path + [existing_name]

        if "properties" in existing and "properties" in defined:
            defined_properties = {p["name"]: p for p in defined.get("properties", [])}
            existing_properties = {
                p.get("api_name", p["name"]): p for p in existing.get("properties", [])
            }
            new_properties_names = (
                defined_properties.keys() - existing_properties.keys()
            )
            undefined_property_names = (
                existing_properties.keys() - defined_properties.keys()
            )

            for name in set(undefined_property_names):
                if provider_version == "ga":
                    undefined_property_names.remove(name)
                    if existing_properties[name].get("min_version", "ga") == "beta":
                        logging.info(
                            "keeping beta field %s from definition of %s in ga ",
                            name,
                            ".".join(path),
                        )
                    else:
                        if existing_properties.get("min_version", "") != "beta":
                            existing_properties["min_version"] = "beta"
                            logging.info(
                                "marking field %s from definition of %s as beta",
                                name,
                                ".".join(path),
                            )

                else:
                    logging.warning(
                        "removing field %s from definition of %s", name, ".".join(path)
                    )

            properties = list(
                filter(
                    lambda p: p.get("api_name", p["name"])
                    not in undefined_property_names,
                    existing["properties"],
                )
            )

            for name in new_properties_names:
                logging.info(
                    "adding %s as %s field to definition of %s",
                    name,
                    provider_version,
                    ".".join(path),
                )
                if provider_version == "beta":
                    defined_properties[name]["min_version"] = "beta"
                properties.append(defined_properties[name])

            existing["properties"] = properties

            for i, current_property in enumerate(existing["properties"]):
                property_name = current_property.get(
                    "api_name", current_property["name"]
                )
                if property_name in defined_properties:
                    defined_property = defined_properties.get(property_name)
                    Field.merge_fields(
                        current_property, defined_property, path, provider_version
                    )
        else:
            if defined.tag.value != existing.tag.value:
                logging.warning(
                    "mismatch of type on field %s, existing tag %s and defined %s",
                    ".".join(path),
                    existing.tag.value,
                    defined.tag.value,
                )
                return
            if defined.tag.value == "!ruby/object:Api::Type::Array":
                Field.merge_fields(
                    existing["item_type"], defined["item_type"], path, provider_version
                )


class Resource(CommentedMap):
    """
    represents a Magic Module resource definition.
    """

    def __init__(self, other: dict):
        super().__init__(self)
        self.tag.value = "!ruby/object:Api::Resource"
        self.update(other)

    @staticmethod
    def load(path: str) -> "Resource":

        with open(path, "r") as file:
            result = yaml.load(file)

        if (
            not isinstance(result, CommentedMap)
            or result.tag.value != "!ruby/object:Api::Resource"
        ):
            raise ValueError(
                f"{path} does not contain a Magic Module object:Api::Resource"
            )

        return Resource(result)

    @staticmethod
    def merge_resources(
        existing: Union["Resource"], defined: Union["Resource"], provider_version: str
    ):

        """
        merges the `defined` resource definition into the `existing` resource definition,
        """
        assert existing is not None
        assert defined is not None
        assert provider_version in ["ga", "beta"]

        Field.merge_fields(existing, defined, [], provider_version)


yaml.representer.add_representer(Resource, RoundTripRepresenter.represent_dict)
yaml.representer.add_representer(Field, RoundTripRepresenter.represent_dict)
yaml.representer.add_representer(Product, RoundTripRepresenter.represent_dict)
